/*
Copyright (c) 1999-2007, United States Government, as represented by
the Administrator for The National Aeronautics and Space Administration.
All rights reserved.
*/
package gov.nasa.gsfc.drl.rtstps.core.output.rdr;

import gov.nasa.gsfc.drl.rtstps.core.RtStpsException;
import gov.nasa.gsfc.drl.rtstps.core.ccsds.Packet;

import java.io.File;
import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import ncsa.hdf.hdf5lib.H5;
import ncsa.hdf.hdf5lib.HDF5Constants;
import ncsa.hdf.hdf5lib.exceptions.HDF5LibraryException;

/**
 * Create an HDF file with basics of the RDR structure.  Provide some methods to further
 * build out the RDR.
 * 
 *
 */
public class RDRCreateFile {
	private int hdfFile;
	private int propertyId;
	private final static int xmlSize = 1536;  // official size per RDR...
	private AllData allData;
	private DataProducts dataProducts;
	
	private PDSDate firstDateTime = null;
	private PDSDate lastDateTime = null;
	
	private SpacecraftDiaryRDR_MultipleGranule spacecraftDiaryRDR_MultipleGranule = null;  //used to be a special case
	
	// all other RDRs go here
	private HashMap<RDRName, RDR> rdrs = new HashMap<RDRName, RDR>();

	//private WeightedRDRList rdrs = new WeightedRDRList();
	
	public static boolean timeBaseCalculation = false;
	
	private Stats stats = null;
	
	
	/**
	 * Open the specified file, wipe out a file of the same name if it exists
	 * A user block space is created based on the rdrCount given.
	 * @param destPath the path of interest
	 * @param filename the file name of interest
	 * @param rdrCount the number of RDRs that will go in this file, used to create the space for the user block
	 * @throws RtStpsException wraps any HDF exceptions
	 */
	public RDRCreateFile(String destPath, String filename, int rdrCount) throws RtStpsException {
		
		int userBlockSize = xmlSize * rdrCount;
		
		// computes the next power of 2 the xml will fit in for the user block
		// which must be a power of two. 
		userBlockSize--;
		userBlockSize = (userBlockSize >>> 1) | userBlockSize;
		userBlockSize = (userBlockSize >>> 2) | userBlockSize;
		userBlockSize = (userBlockSize >>> 4) | userBlockSize;
		userBlockSize = (userBlockSize >>> 8) | userBlockSize;
		userBlockSize = (userBlockSize >>> 16) | userBlockSize;
		userBlockSize++; // userBlockSize is now the next highest power of 2.
		
		// set the user block size
		try {
			propertyId = H5.H5Pcreate (HDF5Constants.H5P_FILE_CREATE);
			H5.H5Pset_userblock(propertyId, userBlockSize);
		} catch (HDF5LibraryException e) {
			throw new RtStpsException(e);
		}
		String pathFilename = destPath + File.separator + filename;
		
		// Use java to check that the directory is there because the HDF
		// library has such nebulous error codes and messages.
		// In order for this to be true it must be a directory AND exist
		File dir = new File(destPath);
		File currentDir = new File(".");
		try {

			if (dir.exists() == false) {

				throw new RtStpsException("RDR write directory [" + destPath
						+ "] does not exist. Current directory ["
						+ currentDir.getCanonicalPath() + "]");

			}

			if (dir.isDirectory() == false) {
				throw new RtStpsException("RDR write directory [" + destPath
						+ "] is not a directory. Current directory ["
						+ currentDir.getCanonicalPath() + "]");
			}
		} catch (IOException e) {
			throw new RtStpsException(e);
		}
		
		//System.out.println("Filename: " + pathFilename);
		
		/* Create a new file. If file exists its contents will be overwritten. */
		try {
			hdfFile = H5.H5Fcreate (pathFilename, HDF5Constants.H5F_ACC_TRUNC, propertyId, HDF5Constants.H5P_DEFAULT);
		} catch (HDF5LibraryException e) {
			throw new RtStpsException(e);
		}
		
		//System.out.println("hdfFile = " + hdfFile);
		
		// create the AllData portion only i.e. "/All_Data"
		allData = new AllData(hdfFile);
		
		// create the DataProducts portion only i.e. "/Data_Products"
		dataProducts = new DataProducts(hdfFile);

	}
	
	public RDRCreateFile(Stats stats, String destPath, String filename, int rdrCount) throws RtStpsException {
	
		this.stats = stats;
		
		int userBlockSize = xmlSize * rdrCount;
		
		// computes the next power of 2 the xml will fit in for the user block
		// which must be a power of two. 
		userBlockSize--;
		userBlockSize = (userBlockSize >>> 1) | userBlockSize;
		userBlockSize = (userBlockSize >>> 2) | userBlockSize;
		userBlockSize = (userBlockSize >>> 4) | userBlockSize;
		userBlockSize = (userBlockSize >>> 8) | userBlockSize;
		userBlockSize = (userBlockSize >>> 16) | userBlockSize;
		userBlockSize++; // userBlockSize is now the next highest power of 2.
		
		// set the user block size
		try {
			propertyId = H5.H5Pcreate (HDF5Constants.H5P_FILE_CREATE);
			H5.H5Pset_userblock(propertyId, userBlockSize);
		} catch (HDF5LibraryException e) {
			throw new RtStpsException(e);
		}
		
		String pathFilename = destPath + File.separator + filename; 
		
		// Use java to check that the directory is there because the HDF
		// library has such nebulous error codes and messages.
		// In order for this to be true it must be a directory AND exist
		File dir = new File(destPath);
		File currentDir = new File(".");
		try {

			if (dir.exists() == false) {

				throw new RtStpsException("RDR write directory [" + destPath
						+ "] does not exist. Current directory ["
						+ currentDir.getCanonicalPath() + "]");

			}

			if (dir.isDirectory() == false) {
				throw new RtStpsException("RDR write directory [" + destPath
						+ "] is not a directory. Current directory ["
						+ currentDir.getCanonicalPath() + "]");
			}
		} catch (IOException e) {
			throw new RtStpsException(e);
		}
		
		
		/* Create a new file. If file exists its contents will be overwritten. */
		try {
			hdfFile = H5.H5Fcreate (pathFilename, HDF5Constants.H5F_ACC_TRUNC, propertyId, HDF5Constants.H5P_DEFAULT);
		} catch (HDF5LibraryException e) {
			throw new RtStpsException(e);
		}
		
		//System.out.println("hdfFile = " + hdfFile);
		
		// create the AllData portion only i.e. "/All_Data"
		allData = new AllData(hdfFile);
		
		// create the DataProducts portion only i.e. "/Data_Products"
		dataProducts = new DataProducts(hdfFile);

	}
	/**
	 * Get the RDR from the pool of created RDRs, or create a new one if its new
	 * @param anRdrName an RDR name as an {@link RDRName}
	 * @return an {@link RDR}
	 * @throws RtStpsException wraps an any HDF files
	 */
	private RDR getRDR(RDRName anRdrName) throws RtStpsException {
		
		RDR rdr = rdrs.get(anRdrName);
		
		if (rdr == null) {
			//rdr = new RDR(anRdrName, allData, dataProducts, FixedDomainDescription.drl);
			
			rdr = RDRFactory.createRDR(stats, anRdrName, allData, dataProducts, FixedDomainDescription.drl);
			
			rdrs.put(anRdrName, rdr);
		}
		
		return rdr;
	}
	
	/**
	 * Return the {@link AllData} object which is created through this class
	 * @return the AllData object
	 */
	public AllData getAllData() {
		return allData;
	}
	
	/**
	 * Return the {@link DataProducts} object which is created through this class
	 * @return the DataProducts object
	 */
	public DataProducts getDataProducts() {
		return dataProducts;
	}
	
	
	/**
	 * Close the HDF file completely, including any open RDRs, AllData and DataProducts
	 * @throws RtStpsException 
	 */
	public void close() throws RtStpsException {
		
		// first close all the normal RDRs
		for (RDR rdr : rdrs.values()) {
			rdr.close();
		}
		
		// then close the spacecraft diary RDR if it is being
		// processed, since it is a special case and depends on 
		// the others
		if (spacecraftDiaryRDR_MultipleGranule != null) {
			//try {
			spacecraftDiaryRDR_MultipleGranule.close(); //rdrs.values());
			//} catch (HDF5Exception e) {
			//	throw new RtStpsException(e);
			//}
			
			// Finally ADD the S/C Diary RDR to the RDR list
			// this is needed by the routine that builds the file name
			rdrs.put(RDRName.NPP_Ephemeris_and_Attitude, spacecraftDiaryRDR_MultipleGranule);
		}
		allData.close();
		dataProducts.close();
		
		try {
			H5.H5Pclose(propertyId);
			H5.H5Fclose(hdfFile);
		} catch (HDF5LibraryException e) {
			throw new RtStpsException(e);
		}
	}

	/**
	 * Return a list of the RDRs created by this class
	 * @return a <code>List</code> of {@link RDR}s
	 */
	public List<RDR> getRDRs() {
		List<RDR> rdrList = new LinkedList<RDR>(rdrs.values());
		return rdrList;
		//System.out.println("rdrList size: " + rdrList.size());
		//return rdrs.values();
	}

	/**
	 * Get the first time of the first packet with a time stamp given to this class
	 * @return a {@link PDSDate} from that packet
	 */
	public PDSDate getFirstDateTime() {
		return firstDateTime;
	}

	/**
	 * Get the last time of the last packet with a time stamp given to this class
	 * @return a {@link PDSDate} from that packet
	 */
	public PDSDate getLastDataTime() {
		return lastDateTime;
	}

	/**
	 * Put a packet into this RDR file
	 * @param p a {@link Packet}
	 * @throws RtStpsException wraps any HDF exception
	 */
	public void put(Packet p) throws RtStpsException {
		
		if (p.isFirstPacketInSequence() || p.isStandalonePacket()) {
			if (firstDateTime == null) {
				firstDateTime = new PDSDate(p.getTimeStamp(8));
			}
			lastDateTime = new PDSDate(p.getTimeStamp(8));
		}
		RDRName an_rdrName = RDRName.fromAppId(p.getApplicationId());
		//System.out.println("an_rdrName " + an_rdrName.getRDRStringName() + " toString: " + an_rdrName.toString());
		
		if (an_rdrName == null) {
			if (stats == null) {
				System.out.println("Unsupported Packet w/App ID[" + p.getApplicationId() + "] -- no matching RDR, dropping packet...");
			} else {
				stats.drop.value++;
			}
		} else if (an_rdrName == RDRName.NPP_Ephemeris_and_Attitude) { 
			
			
			// FIXME spacecraft diary used to be a special case but is not one any longer...
			
			//System.out.println("Special case of Attitude and Ephemeris");
			
			// first if the diary RDR has not been created, create one
			if (spacecraftDiaryRDR_MultipleGranule == null) {
				spacecraftDiaryRDR_MultipleGranule = new SpacecraftDiaryRDR_MultipleGranule(stats, allData, dataProducts, FixedDomainDescription.drl);
			}
			// then put the packet
			spacecraftDiaryRDR_MultipleGranule.put(p);
		} else {
			// all others

			//System.out.println("APID: " + p.getApplicationId() + " RDRName: " + an_rdrName.getRDRStringName());
			RDR rdr = this.getRDR(an_rdrName);
			rdr.put(p);	
		}
	}

	public void writeAttributes(Date creationDateAndTime, 
								Origin distributor, 
								MissionName missionName, 
								Origin datasetSource, 
								PlatformShortName platformShortname) throws RtStpsException {
		StringBuffer dsb = TimeFormat.formatDate(creationDateAndTime);
		StringBuffer tsb = TimeFormat.formatTime(creationDateAndTime);
		
		HDFAttribute.writeString(hdfFile, "Distributor", distributor.toString());
		HDFAttribute.writeString(hdfFile, "Mission_Name", missionName.toString());
		HDFAttribute.writeString(hdfFile, "N_Dataset_Source", datasetSource.toString());
		HDFAttribute.writeString(hdfFile, "N_HDF_Creation_Date", dsb.toString());
		HDFAttribute.writeString(hdfFile, "N_HDF_Creation_Time", tsb.toString());
		HDFAttribute.writeString(hdfFile, "Platform_Short_Name", platformShortname.toString());
	}
	

}
